#include <utility>
#include <cstdio>
#include <vector>
#include <queue>
#include <map>
#include <fcntl.h>
#include <sys/poll.h>

#include "buffer.hh"

class buffer
{
    std::queue<string> buffers;
    unsigned frontptr;
public:
    buffer() : frontptr(0)
    {
    }
    buffer(const string& data) : frontptr(0)
    {
        PushBack(data);
    }
    
    void PushBack(const string& data)
    {
/* - don't use - string append is not very fast.
        if(!buffers.empty())
            buffers.back() += data;
        else
*/
            buffers.push(data);
    }
    void PopFront(unsigned n)
    {
        while(n > 0 && !buffers.empty())
        {
            unsigned c = GetFrontSize();
            if(n < c)
            {
                frontptr += n;
                c = n; // ate "n" bytes
            }
            else
            {
                buffers.pop();
                frontptr = 0;
                // ate everything that was left in the front buf
            }
            n -= c;
        }
    }
    unsigned GetFrontSize() const
    {
        if(buffers.empty()) return 0;
        return buffers.front().size() - frontptr;
    }
    const char *GetFrontPtr() const
    {
        return buffers.front().data() + frontptr;
    }
    
    unsigned GetBufSize() const
    {
        return buffers.empty()
             ? 0
             : buffers.front().size() - frontptr
             + buffers.back().size() * (buffers.size() - 1);
    }
};

static std::map<int, buffer> buffers;

void AddWriteTask(FILE *fp, const string& data)
{
    int fd = fileno(fp);
    
    std::map<int, buffer>::iterator i;
    
    i = buffers.lower_bound(fd);
    if(i == buffers.end() || i->first != fd)
    {
        i = buffers.insert(i, std::make_pair(fd, data));
        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
    }
    else
    {
        i->second.PushBack(data);
    }
}

static void MoveBuffer(int fd, buffer& buf)
{
    unsigned count = buf.GetFrontSize();
    if(!count) return;

    const char *p = buf.GetFrontPtr();
    
    int n = write(fd, p, count);
    if(n > 0)
    {
        buf.PopFront(n);
        // retry!
        MoveBuffer(fd, buf);
    }
}

void MoveBuffers()
{
    std::map<int, buffer>::iterator i;
    for(i = buffers.begin(); i != buffers.end(); ++i)
    {
        MoveBuffer(i->first, i->second);
    }
}

void FlushBuffers()
{
    std::map<int, buffer>::iterator i,j;
    for(i = buffers.begin(); i != buffers.end(); i=j)
    {
        j=i; ++j;
        
        int fd = i->first;

        fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) &~ O_NONBLOCK);

        buffer& buf = i->second;
        
        for(;;)
        {
            unsigned count = buf.GetFrontSize();
            if(!count) break;
            
            const char *p = buf.GetFrontPtr();
            
            int n = write(fd, p, count);
            if(n > 0)
            {
                buf.PopFront(n);
            }
        }
        buffers.erase(i);
    }
}

const string GetBufferStatus()
{
    unsigned num_bytes = 0;
    
    std::map<int, buffer>::const_iterator i;
    for(i = buffers.begin(); i != buffers.end(); ++i)
    {
        num_bytes += i->second.GetBufSize();
    }
    
    char Buf[64];
    sprintf(Buf, "%u kB", num_bytes / 1024);
    return Buf;
}

bool BufferElapse(double nseconds)
{
    std::vector<pollfd> polls;

    std::map<int, buffer>::const_iterator i;
    for(i = buffers.begin(); i != buffers.end(); ++i)
    {
        if(!i->second.GetFrontSize()) continue;
        
        pollfd tmp = { i->first, POLLOUT, 0 };
        polls.push_back(tmp);
    }
    
    if(polls.empty())
    {
        usleep((unsigned)(nseconds * 1e6));
        return false;
    }
    
    int timeout = (int)(nseconds * 1000);
    if(!timeout) timeout=1;
    
    int res = poll(&polls[0], polls.size(), timeout);
    if(!res) return true;
    
    for(unsigned a=0; a<polls.size(); ++a)
        if(polls[a].revents & POLLOUT)
        {
            int fd = polls[a].fd;
            MoveBuffer(fd, buffers[fd]);
        }

    return false;
}
